@@ -103,6 +103,17 @@ class TestFasterCSVInterface < Test::Unit::TestCase
       assert_equal(nil, csv.shift)
     end
   end
+
+  def test_long_line # ruby's regex parser may have problems with long rows
+    File.unlink(@path)
+
+    long_field_length = 2800
+    File.open(@path, "w") do |file|
+      file << "1\t2\t#{'3' * long_field_length}\r\n"
+    end
+    @expected = [%w{1 2} + ['3' * long_field_length]]
+    test_shift
+  end
   
   ### Test Write Interface ###
 
@@ -161,7 +172,6 @@ class TestFasterCSVInterface < Test::Unit::TestCase
 
     lines = [{:a => 1, :b => 2, :c => 3}, {:a => 4, :b => 5, :c => 6}]
     FasterCSV.open( @path, "w", :headers           => true,
-                                :converters        => :all,
                                 :header_converters => :symbol ) do |csv|
       csv << lines.first.keys
       lines.each { |line| csv << line }
@@ -172,6 +182,75 @@ class TestFasterCSVInterface < Test::Unit::TestCase
       csv.each { |line| assert_equal(lines.shift, line.to_hash) }
     end
   end
+
+  def test_write_hash_with_headers_array
+    File.unlink(@path)
+
+    lines = [{:a => 1, :b => 2, :c => 3}, {:a => 4, :b => 5, :c => 6}]
+    FasterCSV.open(@path, "w", :headers => [:b, :a, :c]) do |csv|
+      lines.each { |line| csv << line }
+    end
+
+    # test writing fields in the correct order
+    File.open(@path, "r") do |f|
+      assert_equal("2,1,3", f.gets.strip)
+      assert_equal("5,4,6", f.gets.strip)
+    end
+
+    # test reading CSV with headers
+    FasterCSV.open( @path, "r", :headers    => [:b, :a, :c],
+                                :converters => :all ) do |csv|
+      csv.each { |line| assert_equal(lines.shift, line.to_hash) }
+    end
+  end
+
+  def test_write_hash_with_headers_string
+    File.unlink(@path)
+
+    lines = [{"a" => 1, "b" => 2, "c" => 3}, {"a" => 4, "b" => 5, "c" => 6}]
+    FasterCSV.open( @path, "w", :headers => "b|a|c",
+                                :col_sep => "|" ) do |csv|
+      lines.each { |line| csv << line }
+    end
+
+    # test writing fields in the correct order
+    File.open(@path, "r") do |f|
+      assert_equal("2|1|3", f.gets.strip)
+      assert_equal("5|4|6", f.gets.strip)
+    end
+
+    # test reading CSV with headers
+    FasterCSV.open( @path, "r", :headers    => "b|a|c",
+                                :col_sep    => "|",
+                                :converters => :all ) do |csv|
+      csv.each { |line| assert_equal(lines.shift, line.to_hash) }
+    end
+  end
+  
+  def test_write_headers
+    File.unlink(@path)
+
+    lines = [{"a" => 1, "b" => 2, "c" => 3}, {"a" => 4, "b" => 5, "c" => 6}]
+    FasterCSV.open( @path, "w", :headers       => "b|a|c",
+                                :write_headers => true,
+                                :col_sep       => "|" ) do |csv|
+      lines.each { |line| csv << line }
+    end
+
+    # test writing fields in the correct order
+    File.open(@path, "r") do |f|
+      assert_equal("b|a|c", f.gets.strip)
+      assert_equal("2|1|3", f.gets.strip)
+      assert_equal("5|4|6", f.gets.strip)
+    end
+
+    # test reading CSV with headers
+    FasterCSV.open( @path, "r", :headers    => true,
+                                :col_sep    => "|",
+                                :converters => :all ) do |csv|
+      csv.each { |line| assert_equal(lines.shift, line.to_hash) }
+    end
+  end
   
   def test_append  # aliased add_row() and puts()
     File.unlink(@path)
